#ifndef _STRINGHEAP_H_
#define _STRINGHEAP_H_

#include "Buffer.h"
#include "Basetypes.h"
#include "Memory.h"

typedef int	StringHandle;
enum {
	InvalidStringHandle = (StringHandle)-1,
	EmptyStringHandle = (StringHandle)-2,
};

template<typename T> class TStringHeap {
private:
	enum {
		StringCache_Elemsize = sizeof(T),
		StringCache_ArrayAlign = 256,
		StringCache_BufferAlign = 1024*1024,
		StringCache_UnusedOffset = (int)-1,
		StringCache_EmptyStringOffset = (int)-2,
	};
	int					m_FirstEmpty;
	CDynMemoryBuffer	m_Buffer;
	DynArray<int, StringCache_ArrayAlign, StringCache_UnusedOffset>	m_OffsetArray;
	CriticalSectionClass	m_CS;
	StringHandle FindEmpty(){
		int i, c;
		c = m_OffsetArray.GetSize();
		for (i = 0; i < c; i++){
			if (m_OffsetArray[i] == StringCache_UnusedOffset)
				return (StringHandle)i;
		}
		return InvalidStringHandle;
	}
	StringHandle GetEmptyForUse(){
		StringHandle h;
		int i, c;
		h = (StringHandle)m_FirstEmpty;
		if (h != InvalidStringHandle){
			c = m_OffsetArray.GetSize();
			for (i = m_FirstEmpty+1; i < c; i++){
				if (m_OffsetArray[i] == StringCache_UnusedOffset){
					return m_FirstEmpty = i, h;
				}
			}
		}
		m_FirstEmpty = (int)InvalidStringHandle;
		return h;
	}
	int GetLen(const T* str){
		int len;
		len = 0;
		for (len = 0; *str; str++)
			len++;
		return len;
	}
	//Checks if the string is located in this heap
	bool IsHeapString(const T* str){
		return (str >= m_Buffer.GetPtr() && str <= (void*)(m_Buffer.GetBytePtr()+m_Buffer.GetSize()));
	}
	//Returns offset to heap start
	int GetStringOffset(const T* str){
		return (int)(((LPBYTE)str)-m_Buffer.GetBytePtr());
	}
	T	m_EmptyString[1];
public:
	void Lock(){
		m_CS.Enter();
	}
	void Unlock(){
		m_CS.Leave();
	}
	TStringHeap() : m_Buffer(StringCache_BufferAlign){
		m_EmptyString[0] = 0;
		m_FirstEmpty = (int)InvalidStringHandle;
	}
	~TStringHeap(){
		Clear(true);
	}
	void Clear(bool cleardata = false){
		Lock();
		int i, c;
		m_Buffer.Clear(cleardata);
		if (cleardata){
			m_OffsetArray.Expand(0);
			m_FirstEmpty = (int)InvalidStringHandle;
		} else {			
			m_FirstEmpty = 0;
			for (i = 0, c = m_OffsetArray.GetSize(); i < c; i++)
				m_OffsetArray.Set(i, StringCache_UnusedOffset);
		}
		Unlock();
	}
	StringHandle Add(const T* str){
		StringHandle handle;
		int len, offset, heapoffset;
		bool hs;
		if (!str)
			return InvalidStringHandle;
		len = GetLen(str);
		//Special case for empty strings
		if (len == 0)
			return EmptyStringHandle;
		Lock();
		if (hs = IsHeapString(str))
			heapoffset = GetStringOffset(str);

		handle = GetEmptyForUse();
		if (handle == InvalidStringHandle){
			handle = (StringHandle)m_OffsetArray.GetSize();
			if (!m_OffsetArray.Expand(m_OffsetArray.GetSize()+1))
				return Unlock(), InvalidStringHandle;
			m_FirstEmpty =  ((int)handle+1) < m_OffsetArray.GetSize() ? (int)handle+1 : InvalidStringHandle;
		}
		
		offset = m_Buffer.GetSize();


		if (!m_Buffer.Alloc(((len+1)*sizeof(T))+offset)){
			m_OffsetArray[(int)handle] = StringCache_UnusedOffset;
			return InvalidStringHandle;
		}
		if (hs)
			str = (const T*)(m_Buffer.GetBytePtr()+heapoffset);

		m_OffsetArray[(int)handle] = offset;
		memcpy(m_Buffer.GetBytePtr()+offset, str, (len+1)*sizeof(T));
		Unlock();
		return handle;
	}
	StringHandle Update(StringHandle handle, const T* str){
		int len, offset, newlen, sizedelta, i, size, heapoffset;
		const T* old;
		bool hs;
		if (handle == InvalidStringHandle)
			return Add(str);
		Lock();
		if (!str)
			str = m_EmptyString;
		newlen = GetLen(str);
		if (hs = IsHeapString(str))
			heapoffset = GetStringOffset(str);
		
		if (!newlen)
			  return Remove(handle), Unlock(), EmptyStringHandle;

		if (handle == EmptyStringHandle || (offset = m_OffsetArray.Get((int)handle)) == StringCache_UnusedOffset)
			return Unlock(), Add(str);

		old = (const T*)(m_Buffer.GetBytePtr()+offset);
		len = GetLen(old);

		

		size = (int)m_Buffer.GetSize();

		sizedelta = (newlen-len)*sizeof(T);
		if (sizedelta > 0)
			m_Buffer.Alloc(m_Buffer.GetSize()+sizedelta);

		//Adjust by delta if necessary
		if (hs && heapoffset > len+offset)
			heapoffset += sizedelta;

		if (sizedelta != 0){
			i = size-((len-1)*sizeof(T))-offset;
			memmove_s((m_Buffer.GetBytePtr()+offset+sizedelta+len), i, m_Buffer.GetBytePtr()+offset+len, i);

			
			for (i = ((int)handle)+1, size = m_OffsetArray.GetSize(); i < size; i++){
				if (m_OffsetArray[i] != StringCache_UnusedOffset)
					m_OffsetArray[i] += sizedelta;
			}
		}


		//Update string if source is this heap
		if (hs)
			str = (T*)m_Buffer.GetBytePtr()+heapoffset;
		memcpy(m_Buffer.GetBytePtr()+offset, str, (newlen+1)*sizeof(T));
		Unlock();
		return handle;
	}
	void Remove(StringHandle handle){
		Lock();
		int offset;
		int len, size;
		if (handle == EmptyStringHandle ||(offset = m_OffsetArray.Get((int)handle)) == StringCache_UnusedOffset)
			return Unlock();
		len = (GetLen((const T*)(m_Buffer.GetBytePtr()+offset))+1)*sizeof(T);
		size = (int)m_Buffer.GetSize();
		memmove_s((m_Buffer.GetBytePtr()+offset), size-offset, m_Buffer.GetBytePtr()+m_OffsetArray[(int)handle]+len, size-len-offset);
		m_Buffer.SetSize((size_32)(size-len));
		if ((int)handle < m_FirstEmpty || m_FirstEmpty == InvalidStringHandle)
			m_FirstEmpty = (int)handle;
		for (offset = ((int)handle)+1, size = m_OffsetArray.GetSize(); offset < size; offset++){
			if (m_OffsetArray[offset] != StringCache_UnusedOffset)
				m_OffsetArray[offset] -= len;
		}
		m_OffsetArray[(int)handle] = StringCache_UnusedOffset;
		return Unlock();
	}
	const T* Get(StringHandle handle){
		int offset;
		const T* str;
		if (handle == EmptyStringHandle)
			return m_EmptyString;
		Lock();
		if ((offset = m_OffsetArray.Get((int)handle)) == StringCache_UnusedOffset)
			return Unlock(), NULL;
		str = (const T*)(m_Buffer.GetBytePtr()+offset);
		return Unlock(), str;		
	}
	const T* GetSave(StringHandle handle){
		const T* str;
		str = Get(handle);
		return str ? str : m_EmptyString;
	}
	void PreallocBuffer(int size){
		m_Buffer.EnsureSize((size_32)size);
	}
};

typedef TStringHeap<char>		CStringHeapA;
typedef TStringHeap<wchar_t>	CStringHeapW;

template<typename T> class TStringHandle {
protected:
	StringHandle	m_Handle;
	TStringHeap<T>*	m_pHeap;
	static T		s_EmptyString[1];
	static int Compare(const char* a, const char* b){ return _stricmp(a, b); };
	static int Compare(const wchar_t* a, const wchar_t* b){ return _wcsicmp(a, b); };
public:
	TStringHandle(const TStringHandle<T> &b){
		m_pHeap = b.m_pHeap;
		if (m_pHeap)
			m_Handle = m_pHeap->Add(b.Get());
		else
			m_Handle = InvalidStringHandle;
	}
	TStringHandle(TStringHeap<T> &heap) : m_pHeap(&heap), m_Handle(InvalidStringHandle){};
	TStringHandle(TStringHeap<T> *heap) : m_pHeap(heap), m_Handle(InvalidStringHandle){};
	TStringHandle &operator=(const T* _str){
		if (m_pHeap){
			m_Handle = m_pHeap->Update(m_Handle, _str);
		}
		return *this;
	}
	TStringHandle &operator=(TStringHeap<T> *heap){
		if (m_pHeap){
			m_pHeap->Remove(m_Handle);
			m_Handle = InvalidStringHandle;
		}
		m_pHeap = heap;
		return *this;
	}
	TStringHandle &operator=(TStringHeap<T> &heap){
		if (m_pHeap){
			m_pHeap->Remove(m_Handle);
			m_Handle = InvalidStringHandle;
		}
		m_pHeap = &heap;
		return *this;
	}
	__inline operator const T* () const {
		return m_pHeap ? m_pHeap->GetSave(m_Handle) : s_EmptyString;
	}
	const T* Get() const {
		return *this;
	}
	void Clear(){
		if (m_pHeap)
			m_pHeap->Remove(m_Handle);
		m_Handle = InvalidStringHandle;
	}
	bool operator ==(const T* b){
		return Compare(Get(), b) == 0;
	}
	bool operator !=(const T* b){
		return Compare(Get(), b) != 0;
	}
	const T* operator()() const {
		return *this;
	}
};

template<typename T> T TStringHandle<T>::s_EmptyString[1] = { 0 };


typedef TStringHandle<char>		CStringHandleA;
typedef TStringHandle<wchar_t>	CStringHandleW;


#endif //!_STRINGHEAP_H_
